;;;-*- Mode: Lisp; Package: COMMON-LISP-USER -*-

;;   Copyright (C) 2003 Sandia National Laboratories

;; Sandia National Laboratories grants you the rights to distribute
;; and use this software as governed by the terms
;; of the Lisp Lesser GNU Public License
;; (http://opensource.franz.com/preamble.html),
;; known as the LLGPL.

;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; Lesser General Public License for more details.

;;; employee-browser.lisp
;;; March 4, 2003 Shannon Spires <svspire@sandia.gov>

;====Databrowser Example 1: Editable stuff. This uses the high-level API and should be emulated in real programs.

(in-package :cl-user)

(require :DATABROWSER)

(defparameter *employee-data* #( ("Sam Smith"  45 "800-555-1234" t)
                                 ("Jane Jones" 36 "800-555-2345" nil)
                                 ("Fred Apple" 24 "800-555-3456" t)
                                 ("Sue Miller" 51 "800-555-4567" t)))

(defun set-name (value employee-record)
  (setf (first employee-record) value))

(defun set-age (new-age employee-record)
  "New-age comes in as a string, therefore we have to parse it to get it back into numeric form.
   Keeping it numeric ensures that sorting on this column works intuitively."
  (let ((good-age (ignore-errors (read-from-string new-age nil nil))))
    (if (and good-age (numberp good-age))
      (setf (second employee-record) good-age)
      ; Otherwise, we simply don't store the new value anywhere.
      )))

(defun set-phone (value employee-record)
  (setf (third employee-record) value))

(defun set-training (value employee-record)
  (setf (fourth employee-record) value))

(defun employee-browser (&optional (initial-allow-edits nil))
  (let* ((w (make-instance 'color-dialog
              :window-type :document-with-grow
              :window-title "Employee Records"
              :VIEW-SIZE #@(536 355)))
         (databrowser (MAKE-INSTANCE 'brw:collection-databrowser
                        :column-descriptors #((first ; could have said car here
                                               :title "Name" :justification :left :editable t
                                                     :setter set-name
                                                     ; NOTE: since (setf car) is not a function,
                                                     ;       we have to explicitly specify setters here
                                                     )
                                              (second :title "Age" :editable t :setter set-age
                                                      :maxwidth 100)
                                              (third :title "Phone" :editable t
                                                     :minwidth 120
                                                     :setter set-phone)
                                              (fourth :title "Training Done?" :property-type :checkbox
                                                      :editable t 
                                                      :minwidth 110
                                                      :setter set-training))
                        :triangle-space t ; no twist-downs this demo but leave space anyway
                        :view-nick-name 'employee-browser-view
                        :VIEW-SIZE #@(460 224)
                        :VIEW-POSITION #@(40 58)
                        :view-font '("Hoefler Text" 12 :srcor :plain)
                        :draggable-columns t
                        :view-container w)))
    (MAKE-DIALOG-ITEM 'ccl::CHECK-BOX-DIALOG-ITEM-AR #@(41 298) #@(112 20) "Allow edits"
                      #'(lambda (view) (setf (brw:allow-edits databrowser) (check-box-checked-p view)))
                      :VIEW-NICK-NAME 'ALLOW-EDITS-BUTTON
                      :check-box-checked-p initial-allow-edits
                      :view-container w)

    (setf (brw:allow-edits databrowser) initial-allow-edits) ; so the after method will run

    #| Unnecessary
     (MAKE-DIALOG-ITEM 'BUTTON-DIALOG-ITEM #@(401 298) #@(72 20) "Refresh"
                      #'(lambda (view) (declare (ignore view)) (brw:update databrowser nil))
                      :VIEW-NICK-NAME 'REFRESH-BUTTON
                      :DEFAULT-BUTTON NIL
                      :view-container w)
    |#
    (brw:add-items databrowser *employee-data*)
    databrowser
    ))

(employee-browser nil)

; Notice that if you run (employee-browser nil) a second time, your changes will still be there,
;   unless you reevaluate the defparameter for *employee-data* above.

