;-*- Mode: Lisp; Package: CCL -*-

;;;;;;;;;;;;;;;;;;;;;
;;
;; make-ccl-ballon-help.lisp
;;
;; Create balloon help resources for MCL 2.0
;;

;;;;;;;;;;;;;;;;;;;;;
;;
;; Mod History
;;
;; 07/09/91 bill New File
;;

(in-package :ccl)

(eval-when (compile load eval)
  (require :help-manager))

(defvar *balloon-help-specs*)

(defun make-ccl-balloon-help (&optional (specs *balloon-help-specs*))
  (declare (special *balloon-help-specs*))
  (unless (help-resource-file-refnum *application*)
    (let ((refnum (open-resource-file (help-resource-file *application*)
                                      :if-does-not-exist :create)))
      (when refnum (close-resource-file refnum))))
  (unless (help-resource-file-refnum *application*)
    (error "Can't create help resource file."))
  (dolist (spec specs)
    (add-balloon-help spec)))

(defun add-balloon-help-resource (resource type id)
  (use-resource-file (help-resource-file-refnum *application*))
  (delete-resource type id t)
  (add-resource resource type id)  ; :attributes #$resPurgeable)
  (write-resource resource)
  (detach-resource resource)
  (#_DisposHandle resource))

(defun add-balloon-help (spec)
  (if (atom (car spec))            ; a 'STR ' resource
    (destructuring-bind (id string) spec
      (let ((handle (#_NewHandle (1+ (length string)))))
        (%put-string handle string)
        (add-balloon-help-resource handle :|STR | id)))
    ; a 'STR#' resource
    (let* ((id (caar spec))
           (string-count (length (cdr spec)))
           (size (+ 2 string-count (apply #'+ (mapcar #'length (cdr spec)))))
           (handle (#_NewHandle size))
           (offset 2))
      (%hput-word handle string-count)
      (dolist (string (cdr spec))
        (%put-string handle string offset)
        (setq offset (+ offset (length string) 1)))
      (add-balloon-help-resource handle :|STR#| id))))

(defparameter *balloon-help-specs*
  ; "About Macintosh Common Lisp" on the Apple menu.
  '(((1000)
     "Pop up the about box for Macintosh Common Lisp."
     "Pop up the about box for Macintosh Common Lisp.
The diamond next to \"About \" means that MCL is not idle.")

    ; The "File" menu
    (1100 "File menu

Use this menu to open, close & print windows and to quit the application.
Stored as the value of the symbol *FILE-MENU*.")
    ; "New" on the "File" menu
    (1101 "Create a new empty Fred window by calling FRED.")
    ; "Open" on the "File" menu
    (1102 "Open a Fred window to a file by calling EDIT-SELECT-FILE.")
    ; "Open Selected File" on the "File" menu
    ((1103)
     "Open the specified file by calling OPEN-SELECTED-FILE."
     "Open the specified file by calling OPEN-SELECTED-FILE.
Not available because (OPEN-FILE-ITEM-TITLE (FRONT-WINDOW)) returned NIL (there is no selection in the front window).")
    ; "Close" on the "File" menu
    (1104 "Close the front window by calling WINDOW-CLOSE on it.")
    ; "Save" on the "File" menu
    ((1105)
     "Save the front window to its file by calling WINDOW-SAVE."
     "Save the front window to its file.
Not available because the window has no method for WINDOW-SAVE or WINDOW-NEEDS-SAVING-P or WINDOW-NEEDS-SAVING-P returned NIL.")
    ; "Save As" on the "File" menu.
    ((1106)
     "Save the front window to a file you choose by calling WINDOW-SAVE-AS."
     "Save the front window to a file you choose.
Not available because the window has no method for WINDOW-SAVE-AS.")
    ; "Save Copy As" on the "File" menu
    ((1107)
     "Save a copy of the front window to a file you choose by calling WINDOW-SAVE-COPY-AS."
     "Save a copy of the front window to a file you choose.
Not available because the window has no method for WINDOW-SAVE-COPY-AS.")
    ; "Revert" on the "File" menu
    ((1108)
     "Revert the front window to its contents on disk by calling WINDOW-REVERT."
     "Revert the front window to its contents on disk.
Not available because either the window has no method for WINDOW-REVERT or WINDOW-NEEDS-SAVING-P, or WINDOW-FILENAME or WINDOW-NEEDS-SAVING-P returned NIL.")
    ; "Page Setup" on the "File" menu
    (1109 "Setup the printer by calling the function CCL::PRINT-STYLE-DIALOG")
    ; "Print" on the "File" menu
    ((1110)
     "Print the front window by calling WINDOW-HARDCOPY."
     "Print the front window.
Not available because the window has no method for WINDOW-HARDCOPY.")
    ; "Quit" on the "File" menu
    (1111 "Quit MCL by calling the function QUIT.")

    ; The "Edit" menu
    (1200 "Edit menu

Use this menu to perform editing operations on windows and to undo your last action.
Stored as the value of the symbol *EDIT-MENU*.")
    ; "Undo" on the "Edit" menu
    ((1201)
     "Undo the last action on the front window by calling the function UNDO."
     "Undo the last action on the front window by calling the function UNDO.
Not available because either the window has no method for WINDOW-CAN-UNDO-P or WINDOW-CAN-UNDO-P returned NIL.")
    ; "Undo more" on the "Edit" menu
    ((1202)
     "Undo the action before the one you just undid by calling the function UNDO-NEXT."
     "Undo the action before the one undone by the \"Undo\" menu-item.
Not available because either the window has no method for UNDO-NEXT or WINDOW-CAN-UNDO-NEXT-P or WINDOW-CAN-UNDO-NEXT-P returned NIL.")
    ; "Cut", "Copy", "Paste", "Delete", "Select All" on the "Edit" menu
    (1203 "Cut the selection on the front window to the clipboard by calling the function CUT.")
    (1204 "Copy the selection on the front window to the clipboard by calling the function COPY.")
    (1205 "Paste the clipboard into the front window by calling the function PASTE.")
    (1206 "Delete the selection without copying it to the clipboard by calling the function CLEAR.")
    ((1207)
     "Select the entire front window by calling the function SELECT-ALL."
     "Select the entire front window.
Not available because the window has no method for SELECT-ALL.")
    ; "Search" on the "Edit" menu
    (1208 "Bring up the \"String Search\" dialog to search or replace in the front-window.")

    ; The "Eval Menu"
    (1300 "Eval menu

Use this menu to evaluate buffers, compile or load files, or interact with the break loop.
Stored as the value of the symbol *EVAL-MENU*.")
    ; "Eval Selection" on the "Eval" menu
    ((1301)
     "Evaluate the selection or the expression adjacent to the cursor in the front window by calling the function WINDOW-EVAL-SELECTION"
     "Evaluate the selection or the expression adjacent to the cursor in the front window.
Not available because the window has no method for WINDOW-EVAL-SELECTION.")
    ; "Eval Buffer" on the "Eval" menu
    ((1302)
      "Evaluate the front window by calling the function WINDOW-EVAL-WHOLE-BUFFER."
      "Evaluate the front window.
Not available because the window has no method for WINDOW-EVAL-WHOLE-BUFFER.")
    ; "Load" on the "Eval" menu
     (1303 "Load a file that you choose from the standard file dialog.")
     ; "Compile File" on the "Eval" menu
     (1304 "Compile a file that you choose from the standard file dialog.")
     ; "Abort" on the "Eval" menu
     (1305 "Abort the computation in progress or exit from one level of break loop.")
     ; "Break" on the "Eval" menu
     (1306 "Interrupt the computation in progress and enter a break loop.")
     ; "Continue" on the "Eval" menu
     ((1307)
      "Continue from a continuable error."
      "Continue from a continuable error.
Not available because you are not in a break loop or the most recent error is not continuable.")
     ; "Restarts" on the "Eval" menu
     ((1308)
      "Pop up a menu of restarts."
      "Pop up a menu of restarts.
Not available because you are not in a break loop.")


     ; "Tools" menu
     (1400 "Tools menu

Use this menu to bring up one of the MCL development tools.
Stored as the value of the symbol *TOOLS-MENU*.")
     ; "Apropos"
     (1401 "Bring up a dialog which allows you to list symbols containing a string you type by calling the function APROPOS-DIALOG.")
     ; "Documentation"
     (1402 "Query for a symbol, then display its documentation from the \"MCL Help\" file.")
     ; "Edit Definition"
     (1403 "Query for a symbol and find its definition(s).")
     ; "Inspect"
     (1404 "Bring up the \"Inspector Central\" window which provides help and a bit more with the MCL Inspector.")
     ; "List Definitions"
     ((1405)
      "Bring up a dialog containing the names of all symbols defined in the front window."
      "Bring up a dialog containing the names of all symbols defined in the front window.
Not available because the window has no method for CCL::WINDOW-DEFS-DIALOG.")
     ; "Search Files"
     (1406 "Bring up a dialog that will search files for a string.")
     ; "Backtrace"
     ((1407)
      "Bring up a dialog containing a stack backtrace for the current error."
      "Bring up a dialog containing a stack backtrace for the current error.
Not available because you are not in an error context.")
     ; "Fred Commands"
     (1408 "Bring up a window containing the current Fred key bindings.")
     ; "Print Options"
     (1409 "Bring up a dialog allowing you to change Lisp printing parameters.")
     ; "Environment"
     (1410 "Bring up a dialog allowing you to change MCL global switches.")
      
     (1500 "Windows menu

Use this menu to choose a window to bring to the front.
Stored as the value of the symbol *WINDOWS-MENU*.")
     ((1501)
      "Bring this window to the front."
      "Not available because this window is already in the front."
      "Bring this window to the front.
The cross to the left of the window name means that the window needs to be saved.")

    ))

#|
(make-ccl-balloon-help)
|#
     
      