; demo.lisp
;
; Code to put a time limit on MCL's execution.
;
; At startup, users get a dialog telling them they can use it
; for 15 minutes, or enter a password to be able to use it until
; the end of the month.
; If they enter a valid password, it is stored in their preferences
; folder, so that they won't have to type it again.
;
; To generate a password: (make-password year month day), e.g. (make-password 1996 8 1)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Modification History
;;;
;;; $Log: demo.lisp,v $
;;; Revision 1.4  2003/12/08 08:06:43  gtbyers
;;; Use %remove-standard-method-from-containing-gf instead of %remove-method.
;;;
;;; ------- make 4.3/3.4 demos using this version - so don't trip over xchar-up-down bug
;;;         when starting up (use ~D vs. ~d in format calls)
;;;         (what to do wrto patches for the demos?)
;;; 04/19/97 bill  Make it work for the 68K by making all *standard-kernel-method-class*
;;;                be #+ppc-target
;;; -------------  4.1f1
;;; 01/06/97 bill  Add a comment containing the primes that shipped with 4.0.
;;; 10/25/96 bill  Change the primes so that the new demo will need new passwords.
;;; 10/25/96 bill  The comparison of the password-encoded time with the current time moves from
;;;                valid-password-file-p to valid-password-p so that you can't use MCL with an
;;;                expired password.
;;; 10/24/96 bill  Remove the #+ppc-target before the definition of #_LMGetCurApName
;;;                #+ppc-target in front of advice to active-tsp-count & function-to-dll-header.
;;;                #-ppc-target in front of advice to disassemble-lfun.
;;;                This makes the demo work in MCL 3.1.
;;; -------------  4.0
;;; 10/14/96 bill  Stop "Goodbye" dialog from appearing multiple times.
;;;                exit-dialog protects itself against being thrown out of.
;;;                It also shows the application's name instead of "MCL" and
;;;                takes an optional parameter, which if specified and non-zero
;;;                causes it to print a different message.
;;;                Add message-dialog and modal-dialog to unpatchable-symbols.
;;;                Disable disassembly for the LABELS functions that are part of this file.
;;; -------------  4.0f2
;;; 10/08/96 bill  Bind *listener-p* true in the patched make-mcl-listener-process
;;; 10/05/96 bill  (method window-event :before (t)) sets *running-periodic-tasks* to NIL before
;;;                putting up the quit dialog, so that the time limit can't be usurped by
;;;                causing window-close of a windoid to error or by setting *running-periodic-tasks*
;;;                to true.
;;; 09/24/96 bill  demo@digitool.com -> demo-mcl@digitool.com
;;; -------------  4.0b2
;;; 09/19/96 bill  Don't disallow all redefinitions. Only the ones that implement the demo.
;;;                Enable the disassembler if called from active-tsp-count (part of
;;;                the inspector's restart-frame feature).
;;; 09/17/96 bill  advise %save-application-internal to prevent an explicit toplevel-function arg
;;;                from getting around the demo startup.
;;;                :before advice on %event-dispatch does (check-timeout-validity).
;;;                This is harder to remove than the window-event method.
;;;  9/04/96 slh   hide startup-ticks value (it's easy to find and modify the closed-over value)
;;;                catch errors in check-time-expired (changing closed-over a value can cause one)
;;;                TEMP HACK: allow redefining the file dialog functions
;;; 08/26/96 bill  AlanR's ideas: for kernel functions: disable advise, redefinition, disassembly, and callers.
;;;                ExitToShell if quit or finish-quit is redefined.
;;; 08/21/96 bill  Add license text
;;; -------------  4.0b1
;;; 08/16/96 bill  Add an around method on window-event that quits if the periodic task is disabled
;;;                or changed.
;;;                Make the periodic task quit if the around method is changed or removed.
;;;                Put lots of the former top-level functions in a big LABELS to thwart
;;;                easily finding them with APROPOS.
;;; 08/15/96 bill  Remove the *lisp-system-pointer-functions* entry.
;;;                It was redundant with (method toplevel-function :around (t t)).
;;;

#| Evil hacker notes:
(%put-long (%int-to-ptr #x16A) new-tickcount) does extend the time period, but it also
disables mouse events for a period of time
|#

(in-package :ccl)

; This is a constant to make it harder to change
(defconstant *time-since-startup-in-minutes* 15)

(eval-when (:compile-toplevel :execute)

(deftrap "_LMGetCurApName" () 
  (:no-trap (:pointer (:string 255)))
  (:no-trap (%int-to-ptr #$CurApName)))

)  ; end of eval-when

(defun curapname ()
  (%get-string (#_LMGetCurApName)))

(defun exit-dialog (&optional (count 0))
  (unwind-protect
    (let ((app-name (curapname)))
      (message-dialog
       (format nil
               (if (eql count 0)
                 "~A is quitting now.
You will have a chance to save your changes."
                 "Last chance to save your changes.
~A will quit without asking in 2 minutes.")
               app-name)
       :title "Goodbye"))
    (return-from exit-dialog)))

#|
; Used to find the primes below
(defun prime-p (n)
  (or (eql n 2)
      (and (oddp n)
           (let ((divisor 3))
             (loop
               (multiple-value-bind (q r) (floor n divisor)
                 (when (eql r 0) (return (values nil divisor)))
                 (when (<= q divisor) (return t))
                 (incf divisor 2)))))))


(defun find-prime (start)
  (loop
    (when (prime-p start)
      (return start))
    (incf start 2)))
|#

(let* ((*prime-1* 86753461)     ; 86753437 (4.2)             ; 87654337     (commented out numbers are what shipped with 4.0)
       (*prime-2* 213465977)    ; 213465853                 ; 123456791
       (startup-ticks nil)
       (last-tick-count nil)
       (time-expired-item nil)
       (last-time-reported nil)
       (quitting 0)
       (kill-lisp-pointers #'kill-lisp-pointers)
       (restore-lisp-pointers #'restore-lisp-pointers)
       (quit #'quit)
       (finish-quit #'finish-quit)
       toplevel-function-method window-event-method

       ; These two functions are anonymous so people won't see even their names with (apropos "password")
       (valid-password-p #'(lambda (password)
                             ; Returns a universal time or NIL.
                             (ignore-errors
                              (when (stringp password)
                                (let* ((first-dash (position #\- password)))
                                  (when first-dash
                                    (let* ((prefix (subseq password 0 first-dash))
                                           (suffix (subseq password (1+ first-dash)))
                                           (days (mod (radix-26-value prefix) *prime-1*))
                                           (suffix-days (mod (parse-integer (delete #\- suffix)) *prime-2*)))
                                      (when (eql days suffix-days)
                                        (let* ((year (floor days 10000))
                                               (month (floor (mod days 10000) 100))
                                               (day (mod days 100))
                                               (expiration-time (encode-universal-time 0 0 0 day month year)))
                                          (and (> expiration-time (get-universal-time))
                                               expiration-time))))))))))

       (valid-password-file-p #'(lambda ()
                                  (let ((file (password-file)))
                                    (when (probe-file file)
                                      (let ((password nil))
                                        (with-open-file (s file)
                                          (setq password (read-line s)))
                                        (funcall valid-password-p password)))))))

; This function is removed by a *save-exit-function* so it's not in what we ship
(defun make-password (year month day)
  (let* ((value (+ (* 10000 year) (* 100 month) day))
         (multiplier (logxor year month day))
         (prefix-value (+ value (* (+ multiplier 29) *prime-1*)))
         (prefix (radix-26-string prefix-value))
         (suffix-value (+ value (* (+ multiplier 57) *prime-2*)))
         (suffix (hyphenate-string (format nil "~D" suffix-value))))
    (concatenate 'string prefix "-" suffix)))

(defun make-method-kernel (method)
  #+ppc-target
  (change-class method *standard-kernel-method-class*)
  #-ppc-target
  (let ((f (%method-function method)))
    (lfun-attributes f
                     (bitset $lfatr-kernel-bit
                             (lfun-attributes f)))))

(defun kernel-method-p (thing)
  #+ppc-target
  (typep thing *standard-kernel-method-class*)
  #-ppc-target
  (and (typep thing 'method)
       (logbitp $lfatr-kernel-bit (lfun-attributes (%method-function thing)))))


(macrolet ((magic-number () #x52301)
           (get-startup-ticks ()
             `(%get-long (%inc-ptr (%get-ptr startup-ticks 8) (- (magic-number))) 2)))

  (labels
    ((do-startup-dialog ()
       (let ((expiration-time (funcall valid-password-file-p)))
         (if expiration-time
           (make-expiration-windoid expiration-time)
           (let* ((password (modal-dialog (password-dialog)))
                  (expiration-time (and password (funcall valid-password-p password))))
             (if expiration-time
               (progn
                 (unless (show-license t) (really-quit))
                 (write-password-file password)
                 (make-expiration-windoid expiration-time))
               (progn
                 (unless (show-license t) (really-quit))
                 (install-time-limit)        ;  restart 15 minute countdown.
                 (update-minutes-left-windoid)))))))
     
     (minutes-left ()
       (unless startup-ticks
         (let ((holder (#_NewPtr 8)))
           (%put-long holder (#_TickCount) 2)
           (setq startup-ticks (#_NewPtr 14))
           (%put-ptr startup-ticks (%inc-ptr holder (magic-number)) 8)))
       (let ((minutes-since-startup (floor (- (#_TickCount) (get-startup-ticks))
                                           3600)))
         (- *time-since-startup-in-minutes* minutes-since-startup)))
     
     (update-minutes-left-windoid (&optional (minutes-left (minutes-left)))
       (unless (or (not (eql quitting 0)) (eql minutes-left last-time-reported))
         (flet ((expiration-string ()
                  (format nil "Session ends in ~D minute~:P" minutes-left)))
           (declare (dynamic-extent #'expiration-string))
           (when (null time-expired-item)
             (let ((w (make-instance 'windoid
                        :view-size #@(180 20)
                        :window-show nil
                        :view-subviews
                        (list (setq time-expired-item
                                    (make-dialog-item 'static-text-dialog-item
                                                      #@(0 0)
                                                      nil
                                                      (expiration-string)
                                                      nil
                                                      :view-font "Geneva"))))))
               (set-view-size w (view-size time-expired-item))
               (window-show w)))
           (when (wptr time-expired-item)
             (set-dialog-item-text time-expired-item
                                   (if (> minutes-left 0)
                                     (expiration-string)
                                     "Demo session over.")))
           (setq last-time-reported minutes-left))))
     
     (close-minutes-left-windoid ()
       (when (and time-expired-item (wptr time-expired-item))
         (window-close (view-window time-expired-item)))
       (setq time-expired-item nil))
     
     (maybe-check-timeout-validity ()
       (when startup-ticks
         (check-timeout-validity)))
     
     (check-timeout-validity ()
       (let ((task (find-named-periodic-task 'time-limit-task)))
         (unless (and task
                      (eq (ptask.function task) #'check-time-expired)
                      (eq (find-method #'window-event '(:before) (load-time-value `(,(find-class t))) nil)
                          window-event-method)
                      (eq quit (fboundp 'quit))
                      (eq finish-quit (fboundp 'finish-quit)))
           ; Somebody's messing around. Get out of here.
           (#_ExitToShell))))
     
     (exit-dialog-and-quit ()
       (unless (let ((w (front-window))
                     (md *modal-dialog-on-top*))
                 (and md
                      (listp md)
                      (listp (car md))
                      (eq w (caar md))))
         (without-interrupts
          (let ((saved-quitting quitting))
            (when (eql saved-quitting 2)
              (#_ExitToShell))
            (minutes-left)              ; make sure startup-ticks is initialized
            ; This is in case the user somehow aborts out of the exit-dialog call
            (setf (get-startup-ticks) (+ (- (#_TickCount)
                                            (* 60 60 *time-since-startup-in-minutes*))
                                         (* 120 60)))   ; add two minutes
            (setq quitting (1+ saved-quitting))
            (process-interrupt *event-processor*
                               #'(lambda (saved-quitting startup-ticks)
                                   (without-interrupts
                                    (exit-dialog saved-quitting)
                                    ; This gives two minutes after the dialog is dismissed.
                                    (setf (get-startup-ticks) (+ (- (#_TickCount)
                                                                    (* 60 60 *time-since-startup-in-minutes*))
                                                                 (* 120 60)))   ; add two minutes
                                    (really-quit)))
                               saved-quitting startup-ticks)))))
     
     (check-time-expired ()
       (when (nth-value 1 (ignore-errors 
                           (check-timeout-validity)
                           (let ((ticks (#_TickCount)))
                             (if (and last-tick-count (> last-tick-count ticks))
                               (exit-dialog-and-quit)
                               (let ((minutes-left (minutes-left)))
                                 (setq last-tick-count ticks)
                                 (update-minutes-left-windoid minutes-left)
                                 (when (<= minutes-left 0)
                                   (exit-dialog-and-quit)))))))
         (#_ExitToShell)))
     
     (clear-timeout-state ()
       (setq startup-ticks nil
             last-tick-count nil
             time-expired-item nil
             last-time-reported nil
             quitting 0))
     
     (install-time-limit ()
       (clear-timeout-state)
       (without-interrupts
        (%install-periodic-task 'time-limit-task
                                #'check-time-expired
                                (* 5 60)
                                (+ $ptask_draw-flag $ptask_event-dispatch-flag))
        (add-method #'window-event window-event-method)))
     
     (really-quit ()
       (if (and (eq quit (fboundp 'quit))
                (eq finish-quit (fboundp 'finish-quit)))
         (quit)
         (#_ExitToShell)))
     
     )
    
    ; Don't allow the user to get around the startup-dialog by passing
    ; an explicit toplevel-function argument to save-application.
    (advise %save-application-internal
            (destructuring-bind (&key toplevel-function &allow-other-keys) (cddr arglist)
              (setq startup-ticks nil)      ; make sure %event-dispatch check doesn't go off.
              (when toplevel-function
                (setf (getf (cddr arglist) :toplevel-function)
                      #'(lambda ()
                          (block OK
                            (catch-cancel
                              (restart-case
                                (progn
                                  (do-startup-dialog)
                                  (return-from OK))
                                (abort () (cancel))
                                (abort-break () (cancel))))
                            (#_ExitToShell))
                          (funcall toplevel-function))))
              (:do-it))
            :when :around
            :name 'demo)
    
    (setq toplevel-function-method
          
          (defmethod toplevel-function :around (application init-file)
            (declare (ignore application init-file))
            (block OK
              (catch-cancel
                (restart-case
                  (progn
                    (do-startup-dialog)
                    (return-from OK))
                  (abort () (cancel))
                  (abort-break () (cancel))))
              (#_ExitToShell))
            (%remove-standard-method-from-containing-gf toplevel-function-method)
            (call-next-method))
          
          )  ; end of setq

    (make-method-kernel toplevel-function-method)
    
    (without-interrupts
     
     (setq window-event-method
           
           (defmethod window-event :before ((window t))
             (setq *running-periodic-tasks* nil)
             (check-timeout-validity))
           
           )  ; end of setq
     
     (make-method-kernel window-event-method)
     (%remove-standard-method-from-containing-gf window-event-method)
     
     )  ; end of without-interrupts
    
    (advise %event-dispatch
            (maybe-check-timeout-validity)
            :when :before
            :name 'demo)
    
    ; Patching these two functions is likely the easiest thing for somebody to find.
    ; The kill-lisp-pointers code shows how to disable the time limit.
    (let ((*warn-if-redefine* nil)
          (*warn-if-redefine-kernel* nil))
      
      (setf (symbol-function 'kill-lisp-pointers)
            (nfunction kill-lisp-pointers
                       (lambda ()
                         (without-interrupts
                          (%remove-standard-method-from-containing-gf window-event-method)
                          (%remove-periodic-task 'time-limit-task))
                         (funcall kill-lisp-pointers))))
      
      (setf (symbol-function 'restore-lisp-pointers)
            (nfunction 'restore-lisp-pointers
                       (lambda ()
                         (unless (method-generic-function toplevel-function-method)
                           (add-method #'toplevel-function toplevel-function-method))
                         (funcall restore-lisp-pointers))))
      
      ; Patch the greeting. It should really be in a separate function.
      (defmethod toplevel-function ((a lisp-development-system) init-file)
        (startup-ccl init-file)
        (if *single-process-p*                ; not really right, but ...
          (progn
            (%set-toplevel #'(lambda ()
                               (unless *inhibit-greeting* 
                                 (format t "~&Welcome to the demo version of ~A ~A!~%"
                                         (lisp-implementation-type)
                                         (lisp-implementation-version)))
                               (toplevel-loop)))
            (toplevel))
          (make-mcl-listener-process)))
      
      (defun make-mcl-listener-process ()
        (let ((p (make-process *main-listener-process-name* :background-p t
                               :stack-size *listener-process-stackseg-size*)))
          (process-preset p #'(lambda ()
                                (let ((*listener-p* t))
                                  (unless *inhibit-greeting* 
                                    (format t "~&Welcome to the demo version of ~A ~A!~%"
                                            (lisp-implementation-type)
                                            (lisp-implementation-version)))
                                  (toplevel-loop))))
          (process-enable p)
          p))
      
      )  ; end of let
    
    )))  ; end of labels, macrolet, and let*

(defun password-file ()
  (let ((folder (findfolder #$KOnSystemDisk #$kPreferencesFolderType)))
    (make-pathname :directory (pathname-directory folder)
                   :name "MCL Password")))

(defun show-license (&optional quit-if-not-found)
  (let ((dialog (license-dialog)))
    (if (null dialog)
      (when quit-if-not-found
        (message-dialog "Can't find license resource. Quitting.")
        (quit))
      (progn (catch-cancel 
               (return-from show-license (modal-dialog dialog)))
             nil))))

(defun license-dialog (&key (ok-text "Agree") (cancel-text "Disagree"))
  (let* ((got-resource t)
         (license-resource (or (get-resource :|LCNS| 1 t)
                               (progn
                                 (setq got-resource nil)
                                 (car (find :|LCNS| *patched-resources* :key #'(lambda (x) (and (listp x) (cadr x))))))))
         (license (and license-resource
                       (prog1
                         (with-dereferenced-handles ((p license-resource))
                           (%str-from-ptr p (#_GetHandleSize license-resource)))
                         (when got-resource
                           (release-resource license-resource))))))
    (when (or license 
              (setq license "This should have been a license agreement, but the resource could not be found")
              )
      (let* ((text-item (MAKE-INSTANCE
                         'SCROLLING-FRED-VIEW
                         :dialog-item-text license
                         :view-font '("Geneva" 12)
                         :word-wrap-p t
                         :allow-returns nil
                         :H-SCROLLP nil
                         :V-SCROLLP T
                         :WRAP-P T
                         :VIEW-SIZE #@(467 250)
                         :VIEW-POSITION #@(6 7)))
             (w (MAKE-INSTANCE
                  'COLOR-DIALOG
                  :WINDOW-TYPE :MOVABLE-DIALOG
                  :window-show nil
                  :window-title "License Agreement"
                  :VIEW-POSITION '(:top 50)
                  :VIEW-SIZE #@(480 295)
                  :CLOSE-BOX-P
                  NIL
                  :VIEW-FONT
                  '("Chicago" 12 :SRCOR :PLAIN (:COLOR-INDEX 0))
                  :VIEW-SUBVIEWS
                  `(,(MAKE-DIALOG-ITEM
                      'BUTTON-DIALOG-ITEM
                      #@(404 271)
                      #@(66 18)
                      ok-text
                      #'(lambda (item)
                          (declare (ignore item))
                          (return-from-modal-dialog t))
                      :DEFAULT-BUTTON
                      T)
                    ,@(and cancel-text
                           (list (MAKE-DIALOG-ITEM
                                  'BUTTON-DIALOG-ITEM
                                  #@(322 271)
                                  #@(66 18)
                                  cancel-text
                                  #'(lambda (item)
                                      (declare (ignore item))
                                      (return-from-modal-dialog nil))
                                  :DEFAULT-BUTTON
                                  NIL)))
                    ,text-item))))
        (%buffer-set-read-only (fred-buffer text-item) t)
        (set-selection-range text-item 0 0)
        w))))

(let* ((license-text
"IN PLAIN LANGUAGE

DIGITOOL, INC. DEMONSTRATION SOFTWARE LICENSE

You may use the MCL Demo software, make copies of it, and distribute it to other parties as long as:

1.  You include all its accompanying files, documentation, copyright/trademark notices, and usage license.

2.  You do not modify the MCL Demo in any manner or attempt to circumvent its restrictions.

3.  You do not export it to persons or countries embargoed by the United States or on the U.S. Treasuary Department's restricted list.

IN LEGAL LANGUAGE:

DIGITOOL, INC. DEMONSTRATION SOFTWARE LICENSE

PLEASE READ THIS SOFTWARE LICENSE AGREEMENT \"LICENSE\" CAREFULLY BEFORE USING THE DEMONSTRATION SOFTWARE. BY USING THE DEMONSTRATION SOFTWARE, YOU ARE AGREEING TO BE BOUND BY THE TERMS OF THIS LICENSE.  IF YOU DO NOT AGREE TO THE TERMS OF THIS LICENSE, PROMPTLY REFRAIN FROM FURTHER USE OF THIS SOFTWARE.

1. 	License.  The demonstration software and any fonts accompanying this License whether on disk, in read only memory, or on any other media (the \"Digitool Demonstration Software\") are licensed, not sold, to you by Digitool, Inc. (\"Digitool\").  You own the media on which the Digitool Demonstration Software is recorded but Digitool retains title to the Digitool Demonstration Software.  The Digitool Demonstration Software in this package and any copies which this License authorizes you to make are subject to this License.

2. 	Permitted Uses and Restrictions.  This License allows you to install and use the Digitool Demonstration Software on any number of computers at a time.  You may make copies of the Digitool Demonstration Software in machine-readable form.  The copies must include all copyright information contained on the original.  Except as permitted by applicable law and this License, you may not decompile, reverse engineer, disassemble, circumvent a restriction in, modify, rent, lease, or create derivative works from the Digitool Demonstration Software or transmit the Digitool Demonstration Software over a network.  You may, however, distribute or transmit the Digitool Demonstration Software over a network under this License provided you transfer the related documentation, this License and a copy of the Digitool Demonstration Software to parties who agrees to accept the terms of this License.  Your rights under this License will terminate automatically without notice from Digitool if you fail to comply with any term(s) of this License.

3. 	Disclaimer of Warranty on Digitool Demonstration Software.  You expressly acknowledge and agree that use of the Digitool Demonstration Software is at your sole risk.  The Digitool Demonstration Software is provided \"AS IS\" and without warranty of any kind and Digitool EXPRESSLY DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS FOR A PARTICULAR PURPOSE.  DIGITOOL DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THE DIGITOOL DEMONSTRATION SOFTWARE WILL MEET YOUR REQUIREMENTS, OR THAT THE OPERATION OF THE DIGITOOL DEMONSTRATION SOFTWARE WILL BE UNINTERRUPTED OR ERROR-FREE, OR THAT DEFECTS IN THE DIGITOOL DEMONSTRATION SOFTWARE WILL BE CORRECTED.  FURTHERMORE, DIGITOOL DOES NOT WARRANT OR MAKE ANY REPRESENTATIONS REGARDING THE USE OR THE RESULTS OF THE USE OF THE DIGITOOL DEMONSTRATION SOFTWARE OR RELATED DOCUMENTATION IN TERMS OF THEIR CORRECTNESS, ACCURACY, RELIABILITY, OR OTHERWISE.  NO ORAL OR WRITTEN INFORMATION OR ADVICE GIVEN BY DIGITOOL OR AN DIGITOOL AUTHORIZED REPRESENTATIVE SHALL CREATE A WARRANTY OR IN ANY WAY INCREASE THE SCOPE OF THIS WARRANTY.  SHOULD THE DIGITOOL DEMONSTRATION SOFTWARE PROVE DEFECTIVE, YOU (AND NOT DIGITOOL OR AN DIGITOOL AUTHORIZED REPRESENTATIVE) ASSUME THE ENTIRE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.  SOME JURISDICTIONS DO NOT ALLOW THE EXCLUSION OF IMPLIED WARRANTIES, SO THE ABOVE EXCLUSION MAY NOT APPLY TO YOU.

4. 	Limitation of Liability.  UNDER NO CIRCUMSTANCES, INCLUDING NEGLIGENCE, SHALL DIGITOOL BE LIABLE FOR ANY INCIDENTAL, SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES  ARISING OUT OF OR RELATING TO THIS LICENSE.  SOME JURISDICTIONS DO NOT ALLOW THE LIMITATION OF INCIDENTAL OR CONSEQUENTIAL DAMAGES SO THIS LIMITATION MAY NOT APPLY TO YOU.  In no event shall Digitool's total liability to you for all damages exceed the amount paid for this License to the Digitool Demonstration Software.

5. 	Export Law Assurances.  You may not use or otherwise export or reexport the Digitool Demonstration Software except as authorized by United States law and the laws of the jurisdiction in which the Digitool Demonstration Software was obtained.  In particular, but without limitation, the Digitool Demonstration Software may not be used or otherwise exported or reexported (i) into (or to a national or resident of) any United States embargoed country or (ii) to anyone on the U.S. Treasury Department's list of Specially Designated Nationals or the U.S. Department of Commerce's Table of Denial Orders.  By using the Digitool Demonstration Software, you represent and warrant that you are not located in, under control of, or a national or resident of any such country or on any such list.

6. 	Government End Users. If the Digitool Demonstration Software is supplied to the United States Government, the Digitool Demonstration Software is classified as \"restricted computer software\" as defined in clause 52.227-19 of the FAR.  The United States Government's rights to the Digitool Demonstration Software are as provided in clause 52.227-19 of the FAR.

7. 	Controlling Law and Severability.  This License shall be governed by the laws of the United States and the State of Massachusetts.  If for any reason a court of competent jurisdiction finds any provision, or portion thereof, to be unenforceable, the remainder of this License shall continue in full force and effect.

8. 	Complete Agreement.  This License constitutes the entire agreement between the parties with respect to the use of the Digitool Demonstration Software and supersedes all prior or contemporaneous understandings regarding such subject matter.  No amendment to or modification of this License will be binding unless in writing and signed by Digitool.
")
       (handle (#_NewHandle :errchk (length license-text))))
  (with-dereferenced-handles ((p handle))
    (dotimes (i (length license-text))
      (setf (%get-byte p i) (char-code (aref license-text i)))))
  (setq *patched-resources*
        (cons `(,handle :|LCNS| 1)
              (delete :|LCNS| *patched-resources* :key #'(lambda (x) (and (listp x) (cadr x)))))))

(defun make-expiration-windoid (expiration-time)
  (multiple-value-bind (second minute hour day month year) (decode-universal-time expiration-time)
    (declare (ignore second minute hour))
    (let* ((item (make-dialog-item 'static-text-dialog-item
                                   #@(0 0)
                                   nil
                                   (format nil "Demo period ends on ~D/~D/~D"
                                           month
                                           day
                                           (mod year 100))
                                   nil
                                   :view-font "Geneva"))
           (w (make-instance 'windoid
                :view-size #@(150 20)
                :window-show nil
                :view-subviews
                (list item))))
      (set-view-size w (view-size item))
      (window-show w))))

(defun write-password-file (password)
  (let ((file (password-file)))
    (with-open-file (s file :direction :output :if-exists :supersede)
      (write-string password s))))

(defun password-dialog ()
  (MAKE-INSTANCE
    'COLOR-DIALOG
    :WINDOW-TYPE :MOVABLE-DIALOG
    :window-title "Macintosh Common Lisp Demo"
    :window-show nil
    :VIEW-POSITION '(:top 50)
    :VIEW-SIZE #@(397 157)
    :CLOSE-BOX-P NIL
    :VIEW-SUBVIEWS
    (LIST (MAKE-DIALOG-ITEM
           'EDITABLE-TEXT-DIALOG-ITEM
           #@(96 90)
           #@(224 16)
           ""
           'NIL
           :ALLOW-RETURNS NIL
           :view-nick-name :password)
          (MAKE-DIALOG-ITEM
           'STATIC-TEXT-DIALOG-ITEM
           #@(15 90)
           #@(72 16)
           "Password:"
           'NIL)
          (MAKE-DIALOG-ITEM
           'BUTTON-DIALOG-ITEM
           #@(315 126)
           #@(66 18)
           "OK"
           #'(lambda (item)
               (return-from-modal-dialog (dialog-item-text (view-named :password item))))
           :DEFAULT-BUTTON T)
          (MAKE-DIALOG-ITEM
           'STATIC-TEXT-DIALOG-ITEM
           #@(22 11)
           #@(346 68)
           "This demo version of Macintosh Common Lisp is usable for 15 minutes. To get a password to enable it until the end of the month, send email to demo-mcl@digitool.com."
           'NIL))))

(defun radix-26-value (string)
  (let ((value 0))
    (dotimes (i (length string))
      (let* ((char (char-upcase (char string i)))
             (digit (- (char-code char) (char-code #\A))))
        (setq value (+ (* 26 value) digit))))
    value))

(defun radix-26-string (integer)
  (let ((digits nil)
        (q integer)
        r)
    (loop
      (when (eql 0 q) (return))
      (multiple-value-setq (q r) (floor q 26))
      (push (code-char (+ r (char-code #\A))) digits))
    (coerce (or digits "A") 'string)))

(defun hyphenate-string (string)
  (with-input-from-string (in string)
    (with-output-to-string (out)
      (block done
        (loop
          (dotimes (i 4)
            (let ((char (tyi in)))
              (unless char (return-from done))
              (tyo char out)))
          (when (eofp in) (return-from done))
          (tyo #\- out))))))

; Don't make it trivial for people to cons up new passwords
(defun kill-make-password ()
  (fmakunbound 'make-password)
  (fmakunbound 'kill-make-password)
  (setq *save-exit-functions* (delete 'kill-make-password *save-exit-functions*)))

(pushnew 'kill-make-password *save-exit-functions*)

;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; A little more security
;;; Make it harder to redefine, disassemble, or advise things
;;;

(let ((unpatchable-symbols '(redefine-kernel-method
                             redefine-kernel-function
                             encapsulate
                             unadvise-capsules
                             callers
                             function-to-dll-header
                             kernel-thing-p
                             restore-lisp-pointers
                             kill-lisp-pointers
                             %event-dispatch
                             toplevel-function
                             %save-application-internal
                             exit-dialog
                             quit
                             finish-quit
                             message-dialog
                             modal-dialog
                             do-startup-dialog
                             minutes-left
                             update-minutes-left-windoid
                             close-minutes-left-windoid
                             maybe-check-timeout-validity
                             check-timeout-validity
                             exit-dialog-and-quit
                             check-time-expired
                             clear-timeout-state
                             install-time-limit
                             really-quit
                             make-method-kernel
                             kernel-method-p
                             ))
      #+ppc-target
      (ok-to-disassemble nil))

  (labels ((kernel-thing-p (x)
             (cond ((functionp x)
                    (kernel-thing-p (function-name x)))
                   ((symbolp x)
                    (frozen-definition-p x))
                   ((kernel-method-p x) t)
                   ((and (listp x)
                         (eq (car x) 'setf)
                         (listp (cdr x))
                         (null (cddr x)))
                    (kernel-thing-p (setf-function-name (cadr x))))))
           (thing-name (x)
             (cond ((symbolp x) x)
                   ((functionp x) (thing-name (function-name x)))
                   ((typep x 'method)
                    (thing-name (method-name x)))
                   ((and (listp x) (eq (car x) 'setf) (listp (cdr x)))
                    (cadr x))
                   (t x)))
           (unpatchable-thing-p (x)
             (and (kernel-thing-p x) (memq (thing-name x) unpatchable-symbols))))

#+ppc-target
(advise active-tsp-count
        (unwind-protect
          (progn
            (setq ok-to-disassemble t)
            (:do-it))
          (setq ok-to-disassemble nil))
        :when :around
        :name 'demo)

; Don't allow disassembly of kernel functions or methods.
#+ppc-target
(advise function-to-dll-header
        (if (and (not ok-to-disassemble) (kernel-thing-p (car arglist)))
          ; The inspector & disassembler both expect a function to have at least one instruction.
          (function-to-dll-header #'(lambda (&lap 0) (ppc-lap-function nil () (blr))))
          (progn
            (setq ok-to-disassemble nil)
            (:do-it)))
        :when :around
        :name 'demo)

#-ppc-target
(advise disassemble-lfun
        (if (kernel-thing-p (car arglist))
          '((0 CCL::RTS))
          (:do-it))
        :when :around
        :name 'demo)

; disable callers for kernel functions or methods
(advise callers
        (if (kernel-thing-p (car arglist))
          nil
          (:do-it))
        :when :around
        :name 'demo)

(let ((*warn-if-redefine* nil)
      (*warn-if-redefine-kernel* nil))

; Don't allow advise or unadvise on kernel functions or methods
(defun unadvise-capsules (capsules)
  (let (val)
    (dolist (capsule capsules)
      (let ((spec (encapsulation-spec capsule)))
        (when (unpatchable-thing-p spec)
          (loop
            (error "Demo can't unadvise kernel function: ~S" spec)))
        (push (list spec
                    (encapsulation-advice-when capsule)
                    (encapsulation-advice-name capsule))
              val))
      (remove-encapsulation capsule))
    val))

(advise encapsulate
        (let ((fn-spec (first arglist))
              (old-def (second arglist)))
          (if (or (unpatchable-thing-p fn-spec)
                  (unpatchable-thing-p old-def))
            (loop
              (error "Demo can't trace or advise kernel function: ~S" fn-spec))
            (:do-it)))
        :when :around
        :name 'demo)

; Disable *warn-if-redefine-kernel*

(defun redefine-kernel-method (method)
  (when (kernel-method-p method)
    (if (memq (method-name method) unpatchable-symbols)
      (loop
        (error "The method ~S is already defined in the MCL kernel." method))
      (when *warn-if-redefine-kernel*
        (cerror "Replace the definition of ~S."
                "The method ~S is already defined in the MCL kernel." method)))))

(defun redefine-kernel-function (name)
  (when (and
         (frozen-definition-p name)
         (or (lfunp (fboundp name))
             (and (not (consp name)) (macro-function name)))
         (or (and (consp name) (neq (car name) 'setf))
             (let ((pkg (symbol-package (if (consp name) (cadr name) name))))
               (or (eq *common-lisp-package* pkg) (eq *ccl-package* pkg)))))
    (if (memq name unpatchable-symbols)
      (loop
        (error "The function ~S is already defined in the MCL kernel." name))
      (when *warn-if-redefine-kernel*
        (cerror "Replace the definition of ~S."
                "The function ~S is already defined in the MCL kernel." name)
        (unless (consp name)
          (proclaim-inline nil name))))))

; Make all the new stuff defined in this file part of the kernel

(do-all-symbols (s)
  (when (or (fboundp s) (memq s unpatchable-symbols))
    (%symbol-bits s (bitset $sym_fbit_frozen (%symbol-bits s)))))

(make-all-methods-kernel)

)))  ; end of let, labels, let

