/*
File "plbt.c" - PPC LBT dcmd

Copyright 1990-1994 Apple Computer, Inc.
Copyright 1995 Digitool, Inc.


Modification History

 11/07/95 slh   ndcd version
 11/03/95 slh   PPC version
  5/19/95 slh   dcmdScroll initially
  1/20/95 slh	Use LMGetCurrentA5() instead of register file in case called from
                within trap. Note: sould use dcmdSwapWorlds with LM calls.
*/

#include <Types.h>
#include <LowMem.h>
#include <ndcd.h>
#include "lisp.h"
#include "printsupp.h"


#define LFRAME_BYTES			16
#define frame_backptr(f)	(f[0])
#define lframe_fn(f)			(f[1])
#define lframe_link(f)		(f[2])
#define lframe_vsp(f)			(f[3])
#define MULater_framep(f)	(frame_backptr(f) & 1)
#define lframep(f)				((frame_backptr(f) - (ulong)(f)) == LFRAME_BYTES)

void show_lframe( ulong lfun, ulong link, ulong vsp ) {
  PutUHexZ( link, 8 );
	PutSpacesTo(20);
  PutUHexZ( vsp,  8 );
	PutSpacesTo(30);
  print_object( lfun );
}

void walk_stack( ulong *addr, int nframes ) {
	while ( (nframes--) && ((long) addr) > 0x10000L ) {		// works for me!
		PutUHexZ( addr, 8 );
		PutCStr(": ");
		if ( MULater_framep(addr) ) {
		  PutCLine("-Emulator Frame-");
		} else if ( lframep(addr) ) {
			show_lframe( lframe_fn(addr), lframe_link(addr), lframe_vsp(addr) );
		} else {
			PutCLine("-   C Frame    -");
		}
		addr = (ulong *)(frame_backptr(addr) & ~1);
		*((ulong *) (0x0A78+8)) = addr;
	}
#if 0
	PutCStr("Stopped at ");
	PutUHexZ( addr, 8 );
	PutLine();
#endif
}

pascal void CommandEntry( dcmdBlock* paramPtr ) {
	ulong		addr;
	Boolean ok;
	short		ch;
	
	switch (paramPtr->request) {
	
		case dcmdInit:
			break;
		
		case dcmdHelp:
			PutCLine("PLBT [addr]");
			PutCLine("   Displays active functions for stack starting at 'addr'.");
#if 0
      PutCLine("   If no parameter then use the value of register R1.");
      PutCLine("   Due to PM Debugger brain-death, use PLBT * to show subsequent frames.");
#endif
			break;
		
		case dcmdDoIt:
			ch = dcmdPeekAtNextChar();
      if ( ch == '\n' ) {
        OSErr oserr = dcmdReadPowerPCRegister( R1Register, &addr );
				if ( oserr ) {
					PutCStr( "Sorry, OSErr " );
					PutUDec( oserr );
					PutCLine( " occurred.");
					return;
				}
			} else if ( ch == '*' ) {
				walk_stack( *((ulong *) (0x0A78+8)), 2 ); // 2 safe
			} else {
				(void) dcmdGetNextExpression( (long *) &addr, &ok );
				if (ok) {
					PutCLine("Frame     Link      VSP");
					walk_stack( (ulong *) addr, 1 ); // 1 safe
				} else {
					PutCLine("Syntax error");
				}
			}
			break;
	}
} // CommandEntry

// End of plbt.c
