/*
12/01/95 slh  #<ISTRUCT FOO ...> -> #<FOO ...>
11/28/95 slh  print method/class name, specializers
11/21/95 slh  sprint_instance
11/07/95 slh  ndcd version
11/03/95 slh	PPC version
*/

#include <Types.h>
#include <ndcd.h>
#include <put.h>
#include <string.h>
#include "lisp.h"

void sprint_lisp_expression ( ulong addr, int depth );

#define kPrintBase			16
#define kDfltDepth			5
#define kDfltVectElts		4
#define PutSpace()			PutChar(' ')
#define PutX(n)					PutUHexZ( n, 8 )

void addchar( char c ) {
	PutChar( c );
}

void addstring( char *s ) {
	PutCStr( s );
}

void reset_print_buffer( void ) {}

void display_buffer( void ) {
	PutLine();
}

void PutCLine( char *s ) {
	PutCStr(s);
	PutLine();
}

void sprint_long( ulong x, int base, Boolean negp ) {
	if ( negp ) PutChar('-');
	if ( base == 10 ) {
			PutUDec(x);
	} else {
			PutX( x );
	}
}

void sprint_hex( ulong x ) {
	Boolean negp = ((long) x) < 0;
	if ( negp ) x = -x;
	sprint_long( x, 16, negp );
}

void sprint_fixnum( ulong x, int base ) {
	if ( x ) {
		Boolean negp = ((long) x) < 0;
		if ( negp ) x = -x;
		x >>= ft_shift;
		sprint_long( x, base, negp );
	} else {
		PutChar('0');
	}
}

void sprint_dec( ulong x ) {
	Boolean negp = ((long) x) < 0;
	if ( negp ) x = -x;
  sprint_long( x, 10, negp );
}

void sprint_nil( ulong addr ) {
	ulong lnil;
  PutCStr("NIL");
#if 0
	if ( (dcmdReadPowerPCRegister( R2Register, &lnil ) == noErr) &&
			 (addr != lnil) ) {
		PutCStr("?");
		PutX( addr );
	}
#endif
}

void sprint_vect( ulong addr, char *type, Boolean imms, int start, int max_elts, int depth ) {
	int num_elts = obj_size( addr );

	PutCStr("#<");
  if ( type ) {
		PutCStr(type);
		PutSpace();
	}
	if ( max_elts > num_elts ) max_elts = num_elts;
	while ( max_elts-- ) {
		ulong elt = misc_ref(addr, start++);
		
		if ( imms ) {
  		PutX( elt );
		} else {
			sprint_lisp_expression( elt, depth );
		}
		if ( max_elts ) PutSpace();
	}
  PutChar('>');
}

void sprint_sfloat( ulong addr ) {
  sprint_vect( addr, "SFLOAT", true, 1, 1, 0 );
}

void sprint_dfloat( ulong addr ) {
  sprint_vect( addr, "DFLOAT", true, 2, 2, 0 );
}

void sprint_ratio( ulong addr ) {
  sprint_vect( addr, "RATIO", false, 2, 2, 1 );
}

void sprint_complex( ulong addr ) {
  sprint_vect( addr, "COMPLEX", false, 2, 2, 1 );
}

void sprint_bignum( ulong addr) {
  int size = obj_size(addr);
  
  PutCStr("#<BIGNUM/");
  sprint_dec(size);
  PutCStr("-WORDS ");
  PutX(misc_ref(addr, 1));
  if ( size > 2 ) {
  	PutCStr("");
  }
	if ( size > 1 ) {
		PutX(misc_ref(addr, size - 1));
	}
  PutChar('>');
}

void sprint_string( ulong s, Boolean quotedp ) {
  char *p = (char *) misc_data(s);
  int len = obj_size(s);
  
  if ( quotedp ) PutChar('\"');
  while ( len-- ) {
    PutChar(*p++);
  }
  if ( quotedp ) PutChar('\"');
}

Boolean sym_pname_p( ulong sym, char *pstr1 ) {
	int		len1	= *pstr1++ & 0xFF;
	ulong	pname	= symbol_name(sym);
  char *str2	= (char *) misc_data(pname);
  int		len2	= obj_size(pname);

	if ( len1 == len2 ) {
    while ( len1-- ) {
      if ( *pstr1++ != *str2++ ) return false;
    }
    return true;
  }
  return false;
}

void sprint_symbol( ulong s ) {
  sprint_string( symbol_name(s), 0 );
}

void sprint_function( ulong addr, int depth ) {
  ulong nimms  = obj_size(addr);
  ulong lfbits = misc_ref(addr, nimms);
 
	depth = kDfltDepth;		// assume no recursion will occur

  PutCStr("#<FUNCTION ");
  if ( !(lfbits & (1 << lfbit_noname)) ) {
		ulong name = misc_ref(addr, nimms - 1);
    sprint_lisp_expression( name, depth );
  }
	PutCStr(" @ ");
	PutX( addr );
  PutChar('>');
}

void sprint_instance( ulong instance, int depth ) {
	ulong name	= instance_class_name(instance);
	ulong info	= nil;

	depth = kDfltDepth;		// assume no recursion will occur

	PutCStr("#<");
	if ( stagp(name, st_symbol) ) {
		if ( sym_pname_p( name, "\pSTANDARD-METHOD" ) ) {
			PutCStr("METHOD ");
			name	= method_name(instance);
			info	= method_speczers(instance);
		} else if ( sym_pname_p( name, "\pSTANDARD-CLASS" ) ||
		            sym_pname_p( name, "\pBUILT-IN-CLASS" ) ) {
			info = class_name(instance);
		}
	}
	sprint_lisp_expression( name, depth );
	if ( info ) {
		PutSpace();
		sprint_lisp_expression( info, depth );
	}
	PutCStr(" @ ");
	PutX( instance );
	PutChar('>');
}

void sprint_misc( ulong addr, int depth ) {
  int subtag	= obj_subtag(addr);
	int elts		= kDfltVectElts;
  char *type = nil;
  
  switch( subtag ) {
    case st_bignum:		sprint_bignum( addr );
											break;
    case st_ratio:		sprint_ratio( addr );
											break;
    case st_sfloat:		sprint_sfloat( addr );
											break;
    case st_dfloat:		sprint_dfloat( addr );
											break;
    case st_complex:	sprint_complex( addr );
											break;
    case st_symbol:		sprint_symbol( addr );
											break;
		case st_instance: sprint_instance( addr, depth );
											break;
		case st_function:	sprint_function( addr, depth );
											break;
		case st_string:		sprint_string( addr );
											break;
		default:
			// PutX(addr);
			switch( subtag ) {
				case 0x1F: type = "MACPTR"; break;
				case 0x22: type = "CATCH-FRAME"; break;
				case 0x27: type = "DEAD-MACPTR"; break;
				case 0x2F: type = "CODE-VECTOR"; break;
				case 0x32: type = "SGBUF"; break;
				case 0x42: type = "LOCK"; break;
				case 0x4A: type = "HASH-VECTOR"; break;
				case 0x52: type = "POOL"; break;
				case 0x5A: type = "WEAK"; break;
				case 0x62: type = "PACKAGE"; break;
				case 0x6A: type = "MARK"; break;
				case 0x7A: type = "STRUCT"; break;
				case 0x82: ++elts; break;
				case 0x8A: type = "VALUE-CELL"; break;
				case 0xA2: type = "ARRAYH"; break;
				case 0xAA: type = "VECTORH"; break;
				case 0xAF: type = "SFLOAT-VECTOR"; break;
				case 0xB2: type = "SIMPLE-VECTOR"; break;
				case 0xB7: type = "U32-VECTOR"; break;
				case 0xC7: type = "U8-VECTOR"; break;
				case 0xCF: type = "S8-VECTOR"; break;
				case 0xDF: type = "GENERAL-STRING"; break;
				case 0xE7: type = "U16-VECTOR"; break;
				case 0xEF: type = "S16-VECTOR"; break;
				case 0xF7: type = "DFLOAT-VECTOR"; break;
				case 0xFF: type = "BIT-VECTOR"; break;
				default:
					PutCStr("#<SUBTAG=");
					PutUHexZ(subtag, 2);
					break;
			}
			sprint_vect( addr, type, (subtag & 1), 1, elts, depth );
			break;
  }
}

#if 0
void sprint_cons( ulong cell, int depth ) {
  PutChar('(');
  sprint_lisp_expression( CAR(cell), depth );
  PutChar('');
  PutChar(')');
}
#endif

void sprint_cons( ulong cell, int depth ) {
  ulong cdr;
  PutChar ('(');
  while (1) {
    if ( !nullp(cell) ) {
      sprint_lisp_expression( CAR(cell), depth );
      cdr = CDR(cell);
      if ( !nullp(cdr) ) {
        PutSpace();
        if ( consp(cdr) ) {
          cell = cdr; 
	  	  	continue;
        }
        PutChar('.');
        PutSpace();
        sprint_lisp_expression( cdr, depth ) ;
				break;
      }
    }
    break;
  }
  PutChar(')');
}

void sprint_imm( ulong i) {
	char ch;
  switch( i & 0xFF ) {
    case imm_unbound:
	  	PutCStr("#<UNBOUND>");
	  	break;
		case imm_vsp_protect:
			PutCStr("#<VSP-PROTECT>");
			break;
		case imm_char:
			PutCStr("#\\");
			ch = i >> 16;
			switch( ch ) {
				case 0:    PutCStr("Null");			break;
				case ' ':  PutCStr("Space");		break;
				case '\r': PutCStr("Return");		break;
				case '\n': PutCStr("Newline");	break;
				case '\t': PutCStr("Tab");			break;
				default:
					PutChar( i >> 16 );
				break;
			}
			break;
		case imm_illegal:
			PutCStr("#<ILLEGAL>");
			break;
		default:
			PutCStr("#<IMM ");
			PutX(i);
			PutChar('>');
  }
}

void sprint_lisp_expression ( ulong addr, int depth ) {
  if ( depth-- > 0 ) {
    switch( tag(addr) ) {
      case t_efixnum:
      case t_ofixnum:
      	sprint_fixnum( addr, kPrintBase );
				break;
      case t_cons:
      	sprint_cons( addr, depth );
				break;
      case t_imm:
      	sprint_imm( addr );
				break;
			case t_nil:
				sprint_nil( addr );
				break;
			case t_nodevector:
			case t_immvector:
				// Should only appear in subtags, t_misc is used for all vectors.
				PutChar('?');
				// fall twoo
			case t_misc:
				sprint_misc( addr, depth );
				break;
    }
  } else PutChar('#');
}
  
void print_lisp_expression( ulong addr, int depth ) {
  sprint_lisp_expression( addr, depth );
  PutLine();
}

void print_object( ulong value ) {
	print_lisp_expression( value, kDfltDepth );
}


// end
